@@ -6,6 +6,7 @@
 #include "alarm.h"
 #include "filter.h"
 #include "channel.h"
+#include "internal.h"
 
 #include <stdint.h>
 #include <stdio.h>
@@ -95,6 +96,19 @@ struct ct_conf {
 	int poll_kernel_secs;
 	int filter_from_kernelspace;
 	int event_iterations_limit;
+	struct {
+		int error_queue_length;
+	} channelc;
+	struct {
+		int internal_cache_disable;
+		int external_cache_disable;
+	} sync;
+	struct {
+		int events_reliable;
+	} netlink;
+	struct {
+		int commit_steps;
+	} general;
 	struct {
 		int type;
 		int prio;
@@ -165,15 +179,29 @@ struct ct_general_state {
 #define STATE_SYNC(x) state.sync->x
 
 struct ct_sync_state {
-	struct cache *internal; 	/* internal events cache (netlink) */
-	struct cache *external; 	/* external events cache (mcast) */
-
-	struct nfct_handle *commit;
+	struct external_handler *external;
 
 	struct multichannel	*channel;
 	struct nlif_handle	*interface;
 	struct queue *tx_queue;
 
+#define COMMIT_STATE_INACTIVE	0
+#define COMMIT_STATE_MASTER	1
+#define COMMIT_STATE_RELATED	2
+
+	struct {
+		int			state;
+		int			clientfd;
+		struct nfct_handle	*h;
+		struct evfd		*evfd;
+		int			current;
+		struct {
+			int 		ok;
+			int		fail;
+			struct timeval	start;
+		} stats;
+	} commit;
+
 	struct alarm_block		reset_cache_alarm;
 
 	struct sync_mode *sync;		/* sync mode */
@@ -212,18 +240,11 @@ extern union ct_state state;
 extern struct ct_general_state st;
 
 struct ct_mode {
+	struct internal_handler *internal;
 	int (*init)(void);
 	void (*run)(fd_set *readfds);
 	int (*local)(int fd, int type, void *data);
 	void (*kill)(void);
-	void (*dump)(struct nf_conntrack *ct);
-	int (*resync)(enum nf_conntrack_msg_type type,
-		      struct nf_conntrack *ct,
-		      void *data);
-	int (*purge)(void);
-	void (*event_new)(struct nf_conntrack *ct, int origin);
-	void (*event_upd)(struct nf_conntrack *ct, int origin);
-	int (*event_dst)(struct nf_conntrack *ct, int origin);
 };
 
 /* conntrackd modes */
@@ -234,7 +255,6 @@ extern struct ct_mode stats_mode;
 
 /* These live in run.c */
 void killer(int foo);
-void local_handler(int fd, void *data);
 int init(void);
 void run(void);
 
