@@ -1401,7 +1401,8 @@ int main(int argc, char *argv[])
 		
 	case CT_EVENT:
 		if (options & CT_OPT_EVENT_MASK)
-			cth = nfct_open(CONNTRACK, event_mask);
+			cth = nfct_open(CONNTRACK,
+					event_mask & NFCT_ALL_CT_GROUPS);
 		else
 			cth = nfct_open(CONNTRACK, NFCT_ALL_CT_GROUPS);
 
@@ -1444,19 +1445,18 @@ int main(int argc, char *argv[])
 		break;
 	case CT_COUNT: {
 #define NF_CONNTRACK_COUNT_PROC "/proc/sys/net/netfilter/nf_conntrack_count"
-		int fd, count;
-		char buf[strlen("2147483647")];	/* INT_MAX */
-		fd = open(NF_CONNTRACK_COUNT_PROC, O_RDONLY);
-		if (fd == -1) {
+		FILE *fd;
+		int count;
+		fd = fopen(NF_CONNTRACK_COUNT_PROC, "r");
+		if (fd == NULL) {
 			exit_error(OTHER_PROBLEM, "Can't open %s",
 				   NF_CONNTRACK_COUNT_PROC);
 		}
-		if (read(fd, buf, sizeof(buf)) == -1) {
+		if (fscanf(fd, "%d", &count) != 1) {
 			exit_error(OTHER_PROBLEM, "Can't read %s",
 				   NF_CONNTRACK_COUNT_PROC);
 		}
-		close(fd);
-		count = atoi(buf);
+		fclose(fd);
 		printf("%d\n", count);
 		break;
 	}
@@ -1493,7 +1493,7 @@ int main(int argc, char *argv[])
 
 	free_options();
 
-	if (exit_msg[cmd][0]) {
+	if (command && exit_msg[cmd][0]) {
 		fprintf(stderr, "%s v%s (conntrack-tools): ",PROGNAME,VERSION);
 		fprintf(stderr, exit_msg[cmd], counter);
 		if (counter == 0 && !(command & (CT_LIST | EXP_LIST)))
